from typing import cast
import pytest

from idac_sdk import IDACRequestSync, IDACControllerSync, IDACAuthType, IDACRequestType, SessionData
from idac_sdk.errors import SessionDataFileNotFoundError

from xml.parsers.expat import ExpatError

longrun = pytest.mark.skipif("not config.getoption('longrun')")


@pytest.fixture(scope="session")
def test_dev_controller():
    return IDACControllerSync(auth_type=IDACAuthType.NONE)


@pytest.mark.parametrize(
    "session_xml_path, initial_data, test_props",
    [
        (False, {"location": "test.location"}, None),
        (
            "good_session.xml",
            {"recipeName": "nop_v1"},
            {"datacenter": "LON", "id": "551495"},
        ),
    ],
)
def test_session_data_ok(test_data_path, session_xml_path, initial_data, test_props):
    if session_xml_path:
        file = f"{test_data_path}/{session_xml_path}"
    else:
        file = session_xml_path

    sd = SessionData(session_xml_path=file, initial_data=initial_data)
    assert isinstance(sd, SessionData)
    if isinstance(initial_data, dict):
        for k, v in initial_data.items():
            assert sd.has(k), f"Prop {k} not set on SessionData"
            assert sd.get(k) == v, f"Value of prop {k} is different"

    if isinstance(test_props, dict):
        for k, v in test_props.items():
            assert sd.has(k), f"Prop {k} not set on SessionData"
            assert sd.get(k) == v, f"Value of prop {k} is different"


@pytest.mark.parametrize(
    "session_xml_path, initial_data, expected",
    [
        ("bad_session.xml", None, ExpatError),
        ("non_existent.xml", None, SessionDataFileNotFoundError),
    ],
)
def test_session_data_error(test_data_path, session_xml_path, initial_data, expected):
    if session_xml_path:
        file = f"{test_data_path}/{session_xml_path}"
    else:
        file = session_xml_path

    with pytest.raises(expected):
        sd = SessionData(session_xml_path=file, initial_data=initial_data)
        assert isinstance(sd, SessionData)


def test_empty_request():
    p = IDACRequestSync(session_data=SessionData(session_xml_path=False))
    assert isinstance(p, IDACRequestSync)


@pytest.mark.parametrize(
    "session_xml_path, initial_data, request_type",
    [
        pytest.param(
            False,
            {"recipeName": "nop-v1", "recipePath": "common/", "owner": "sdk_test"},
            IDACRequestType.SIMPLE,
            marks=pytest.mark.dependency(name="test_recipe_1"),
        ),
        pytest.param(
            False,
            {
                "recipeName": "hello-world",
                "recipePath": "common/",
                "owner": "sdk_test",
                "datacenter": "lon",
            },
            IDACRequestType.STATELESS,
            marks=pytest.mark.dependency(name="test_recipe_2"),
        ),
    ],
)
def test_recipe(
    request, test_data_path, test_dev_controller, session_xml_path, initial_data, request_type
):
    if session_xml_path:
        file = f"{test_data_path}/{session_xml_path}"
    else:
        file = session_xml_path

    sd = SessionData(initial_data=initial_data, session_xml_path=file)
    req = IDACRequestSync(session_data=sd, controller=test_dev_controller)

    state, redirect = req.create(request_type=request_type)
    assert state.request and state.request.uuid
    request.config.cache.set(f"created_uuid_{sd.get('recipeName')}", state.request.uuid)
    request.config.cache.set(f"redirect_{sd.get('recipeName')}", redirect)
    assert True


@pytest.mark.dependency(depends=["test_recipe_1"])
def test_get_state(request, test_dev_controller):
    req = IDACRequestSync(
        uuid=request.config.cache.get("created_uuid_nop-v1", None),
        controller=test_dev_controller,
        session_data=SessionData(session_xml_path=False),
    )

    state = req.get_state()
    assert state.status


@pytest.mark.dependency(depends=["test_recipe_1"])
def test_extend(request, test_dev_controller):
    req = IDACRequestSync(
        uuid=request.config.cache.get("created_uuid_nop-v1", None),
        controller=test_dev_controller,
        session_data=SessionData(session_xml_path=False),
    )

    req.extend(60)
    assert True


@pytest.mark.dependency(depends=["test_recipe_1"])
def test_cleanup(request, test_dev_controller):
    req = IDACRequestSync(
        uuid=request.config.cache.get("created_uuid_nop-v1", None),
        controller=test_dev_controller,
        session_data=SessionData(session_xml_path=False),
    )

    req.cleanup()
    assert True


@pytest.mark.dependency(depends=["test_recipe_2"])
def test_redirection(request):
    redirect = request.config.cache.get("redirect_hello-world", None)
    uuid = request.config.cache.get("created_uuid_hello-world", None)
    assert isinstance(redirect, str)
    assert uuid in redirect


@pytest.mark.dependency(depends=["test_recipe_2", "test_redirection"])
def test_cleanup_hello_world(request, test_dev_controller):
    req = IDACRequestSync(
        uuid=request.config.cache.get("created_uuid_hello-world", None),
        controller=test_dev_controller,
        session_data=SessionData(session_xml_path=False),
    )

    st = req.wait_for_status(wanted_state=["executed"], max_attempts=5, interval=5)
    assert st and st.status == "executed"

    req.cleanup()
    assert True


@longrun
@pytest.mark.parametrize(
    "session_xml_path, initial_data, request_type, amount",
    [
        pytest.param(
            False,
            {"recipeName": "nop-v1", "recipePath": "common/", "owner": "sdk_test"},
            IDACRequestType.SIMPLE,
            10,
            marks=pytest.mark.dependency(name="test_multiple_1"),
        ),
    ],
)
def test_multiple(
    request,
    test_data_path,
    test_dev_controller,
    session_xml_path,
    initial_data,
    request_type,
    amount,
):
    if session_xml_path:
        file = f"{test_data_path}/{session_xml_path}"
    else:
        file = session_xml_path

    request.config.cache.set(f"multi_amount", amount)

    for idx in range(0, amount):
        cast(dict, initial_data).update({"demo": f"Multi test, request #{idx}"})
        sd = SessionData(initial_data=initial_data, session_xml_path=file)
        req = IDACRequestSync(session_data=sd, controller=test_dev_controller)

        state, redirect = req.create(request_type=request_type)
        assert state.request and state.request.uuid
        request.config.cache.set(f"created_uuid_{idx}_{sd.get('recipeName')}", state.request.uuid)


@longrun
@pytest.mark.dependency(depends=["test_multiple_1"])
def test_multi_awaits(request, test_dev_controller):
    amount = request.config.cache.get("multi_amount", None)

    for idx in range(0, amount):
        uu = request.config.cache.get(f"created_uuid_{idx}_nop-v1", None)
        assert uu
        req = IDACRequestSync(
            uuid=uu,
            controller=test_dev_controller,
            session_data=SessionData(session_xml_path=False),
        )

        st = req.wait_for_status(wanted_state=["executed"], max_attempts=5, interval=5)
        assert st and st.status == "executed"



r"""
Base classes for iDAC Controller & iDAC Request.
Should not be used directly!
"""

import platform
from idac_sdk._version import __version__ as version
from idac_sdk.types import IDACRequestStatus

system = platform.platform()
python = platform.python_implementation() + " " + platform.python_version()

DEFAULT_WAIT_TIMEOUT = 10 * 60  # 10 minutes
DEFAULT_WAIT_INTERVAL = 30  # 30 seconds
DEFAULT_MAX_TRIES = int(DEFAULT_WAIT_TIMEOUT / DEFAULT_WAIT_INTERVAL)
DEFAULT_USER_AGENT = f"iDAC SDK/{version} ({system}; {python})"

REQUEST_ERROR_STATES = [
    IDACRequestStatus.cancelled,
    IDACRequestStatus.error,
    IDACRequestStatus.onboardError,
]
REQUEST_GOOD_STATES = [
    IDACRequestStatus.active,
    IDACRequestStatus.executed,
    IDACRequestStatus.complete,
]



"""
.. include:: ../docs/src/main.md
"""


from idac_sdk.types import IDACAuthType, IDACRequestStatus, IDACRequestType
from idac_sdk.session_data import SessionData
from idac_sdk.asynced.request import IDACRequest as IDACRequestAsync
from idac_sdk.synced.request import IDACRequest as IDACRequestSync
from idac_sdk.asynced.controller import IDACController as IDACControllerAsync
from idac_sdk.synced.controller import IDACController as IDACControllerSync
from idac_sdk.models.vpn_config import VPNConfig, VPNType



from __future__ import annotations
from typing import Optional
from pydantic import BaseModel, Extra

from idac_sdk.models.auth import Auth
from idac_sdk.models.vpn_config import VPNConfig


class Defaults(BaseModel, extra=Extra.allow):
    idac_fqdn: str
    idac_proto: str
    auth: Auth
    api_version: str
    vpn: VPNConfig = VPNConfig()


class iDACConfig(BaseModel, extra=Extra.allow):
    version: str
    defaults: Defaults



from typing import Optional, Tuple, Union
from httpx import Client
from idac_sdk.base.controller import BaseIDACController

from idac_sdk.models.vpn_config import VPNConfig
from idac_sdk.types import IDACAuthType

from idac_sdk.base.controller import BaseIDACController

# from idac_sdk.synced.helpers import sync_method


class IDACController(BaseIDACController):
    __idac_synced_obj__ = True

    def __init__(
        self,
        proto: Optional[str] = None,
        url: Optional[str] = None,
        api_version: Optional[str] = None,
        auth_type: Optional[IDACAuthType] = None,
        auth: Optional[Union[str, Tuple[str, str]]] = None,
        worker_id: Optional[str] = None,
        worker_secret: Optional[str] = None,
        user_agent: Optional[str] = None,
        vpn: Optional[VPNConfig] = None,
    ) -> None:
        """
        IDACController Object

        IDACController describes iDAC controller instance.
        Handles generation of API URLs and Auth Headers required for authentication.

        Args:
            proto (str, optional): `http` or `https`.
            url (str, optional): URL of the controller.
            api_version (str, optional): API version.
            auth_type (IDACAuthType, optional): Type of authentication.
            auth (Union[str, Tuple[str, str]], optional): Authentication data. Usually a token. If
                BASIC auth type is set, can be specified as a tuple in format [username, password]
                which will be automatically encoded in base64. Defaults to None.
            user_agent (Optional[str]): User-Agent string.

        Returns:
            None

        Raises:
            IncorrectControllerProtoError: Raised if incorrect proto provided
            IncorrectControllerURLError: Raised if no URL provided
        """
        super().__init__(
            proto=proto,
            url=url,
            api_version=api_version,
            auth_type=auth_type,
            auth=auth,
            worker_id=worker_id,
            worker_secret=worker_secret,
            user_agent=user_agent,
            vpn=vpn,
        )

    def get_auth_token(
        self, creds: str = "", datacenter: str = "", use_as_auth: bool = True
    ) -> str:
        """Requests auth token from controller. Required for DCLOUD_SESSION and WORKER auth type

        Args:
            creds (str): `creds` token from session.xml dCloud file
            datacenter (str): dCloud datacenter where session runs
            use_as_auth (bool, optional): If True, response will be used as `auth` for
                `auth_headers` & `with_auth` methods. Defaults to True.

        Raises:
            NoAuthTokenInResponse: Raised if no token received in response
            NoAuth: Raised if no authentication provided
        Returns:
            str: Auth token
        """
        prep = self.get_auth_token_pre(creds=creds, datacenter=datacenter)
        if isinstance(prep, str):
            return prep

        # unpack
        (api, headers, data) = prep

        with Client() as client:
            client.headers.update({"User-Agent": self.user_agent})
            response = client.post(api, headers=headers, json=data)
            return self.verify_auth_response(response=response, use_as_auth=use_as_auth)



from pydantic import BaseModel, Extra
from enum import Enum
from typing import Optional


class VPNType(str, Enum):
    none = "none"
    vpod = "vpod"
    explicit = "explicit"
    request = "request"
    secure_repo = "secure_repo"


class VPNParameters(BaseModel, extra=Extra.allow):
    # for type set explicit or request
    host: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None

    # for type set secure_repo
    secure_repo_space: Optional[str] = None
    secure_repo_file: Optional[str] = None
    secure_repo_block: Optional[str] = None
    secure_repo_key: Optional[str] = None


class VPNConfig(BaseModel, extra=Extra.allow):
    type: VPNType = VPNType.none
    params: VPNParameters = VPNParameters()



class SessionDataFileNotFoundError(FileNotFoundError):
    ...


class SessionDataFileNotFoundWarning(Warning):
    ...


class IncorrectControllerProtoError(Exception):
    ...


class IncorrectControllerURLError(Exception):
    ...


class IncorrectSessionData(Exception):
    ...


class NoIdError(Exception):
    ...


class NoControllerError(Exception):
    ...


class NoAuthTokenInResponse(Exception):
    ...


class NoAuth(Exception):
    ...


class BadXmlFileError(Exception):
    ...


class IDACRequestStateError(Exception):
    ...


class IncorrectWantedStateError(Exception):
    ...


class IncorrectMinutesValue(Exception):
    ...


class NoIDACConfig(Warning):
    ...


class OldConfigFoundException(Exception):
    ...


class OldConfigFoundWarning(Warning):
    ...


class UnknownVPNType(Exception):
    ...



from typing import Dict, Optional, Tuple, Union
import json
import httpx
from base64 import b64encode
import warnings
from operator import itemgetter
from idac_sdk.lib._helpers import get_worker_hash
from idac_sdk.models.auth import Auth

from idac_sdk.log import logger
from idac_sdk.config import have_config, load_config
from idac_sdk.models.vpn_config import VPNConfig
from idac_sdk.types import IDACAuthType
from idac_sdk.errors import (
    IncorrectControllerProtoError,
    IncorrectControllerURLError,
    NoAuth,
    NoAuthTokenInResponse,
    NoIDACConfig,
    NoIdError,
)

from . import DEFAULT_USER_AGENT


class BaseIDACController:
    proto: str
    url: str
    api_version: str
    auth: Auth
    user_agent: str
    vpn: VPNConfig

    def __init__(
        self,
        proto: Optional[str] = None,
        url: Optional[str] = None,
        api_version: Optional[str] = None,
        auth_type: Optional[IDACAuthType] = None,
        auth: Optional[Union[str, Tuple[str, str]]] = None,
        worker_id: Optional[str] = None,
        worker_secret: Optional[str] = None,
        user_agent: Optional[str] = None,
        vpn: Optional[VPNConfig] = None,
    ) -> None:
        """
        Base IDACController Object

        IDACController describes iDAC controller instance.
        Handles generation of API URLs and Auth Headers required for authentication.

        Args:
            proto (str, optional): `http` or `https`.
            url (str, optional): URL of the controller.
            api_version (str, optional): API version.
            auth_type (IDACAuthType, optional): Type of authentication.
            auth (Union[str, Tuple[str, str]], optional): Authentication data. Usually a token. If
                BASIC auth type is set, can be specified as a tuple in format [username, password]
                which will be automatically encoded in base64. Defaults to None.
            user_agent (Optional[str]): User-Agent string.

        Returns:
            None

        Raises:
            IncorrectControllerProtoError: Raised if incorrect proto provided
            IncorrectControllerURLError: Raised if no URL provided
        """
        if not have_config():
            warnings.warn("iDAC config not found", NoIDACConfig)

        cfg = load_config()

        self.proto = proto if proto else cfg.defaults.idac_proto
        if self.proto != "http" and self.proto != "https":
            raise IncorrectControllerProtoError(f"Unsupported proto: {proto}")

        self.url = url if url else cfg.defaults.idac_fqdn
        if not self.url:
            raise IncorrectControllerURLError("Controller URL not provided")

        self.user_agent = user_agent if user_agent else DEFAULT_USER_AGENT
        self.api_version = api_version if api_version else cfg.defaults.api_version

        auth_type = auth_type if auth_type else IDACAuthType[cfg.defaults.auth.type]
        params: dict[str, str] = {}
        if (auth and isinstance(auth, str) and auth_type not in [
                IDACAuthType.DCLOUD_SESSION.name,
                IDACAuthType.NONE.name,
        ]):
            params = {"token": auth}
        elif auth_type == IDACAuthType.BASIC.name:
            if isinstance(auth, list):
                params = {
                    "token": b64encode(bytes(f"{auth[0]}:{auth[1]}", "utf-8")).decode("utf-8")
                }
        elif auth_type == IDACAuthType.WORKER.name:
            if worker_id and worker_secret:
                params = {"secret": worker_secret, "client_id": worker_id}
        else:
            params = {}
        self.auth = Auth(type=auth_type.name, params=params)
        self.vpn = vpn or cfg.defaults.vpn

    def api_string(self, api: str) -> str:
        """Generate full URL for an API

        Args:
          api (str): API

        Returns:
          str: Full API URL
        """
        result = f"{self.proto}://{self.url}"
        if not result.endswith("/"):
            result = result + "/"
        return result + f"api/v{self.api_version}/{api}"

    def auth_headers(self) -> Dict[str, str]:
        """Generates Authorization headers required for APIs

        Returns:
          Dict[str, str]: Headers
        """
        if not self.auth or self.auth.type == IDACAuthType.NONE.name:
            return {}

        if not self.auth.params or not "token" in self.auth.params:
            raise NoAuth("No authentication data")

        if self.auth.type == IDACAuthType.BASIC.name:
            return {"Authorization": f"Basic {self.auth.params['token']}"}

        return {"Authorization": f"Bearer {self.auth.params['token']}"}

    def with_auth(self, headers: Dict[str, str]) -> Dict[str, str]:
        """Adds auth headers to `headers` param. Also returns that dictionary.

        Example:
        ```
        h = {"Accept": "application/json"}
        idac_controller.with_auth(h)
        ```

        Args:
          headers (Dict[str, str]): Dictionary where headers should be added

        Returns:
          Dict[str, str]: Updated dictionary
        """
        if self.auth.type == IDACAuthType.NONE.name:
            return headers

        if not isinstance(headers, dict):
            raise Exception("Headers not provided")

        headers.update(self.auth_headers())
        return headers

    def api_create_stateless(self) -> Tuple[str, str]:
        """Generate URL for "Create Stateless Request" API

        Returns:
          Tuple[str, str]: [URL, Method]
        """
        return self.api_string("request/stateless/0"), "GET"

    def api_create_stateful(self) -> Tuple[str, str]:
        """Generate URL for "Create Stateful Request" API

        Returns:
          Tuple[str, str]: [URL, Method]
        """
        return self.api_string("request/stateful/0"), "GET"

    def api_create(self) -> Tuple[str, str]:
        """Generate URL for "Create Request" API

        Returns:
          Tuple[str, str]: [URL, Method]
        """
        return self.api_string("request/0"), "POST"

    def api_get_state(self, id: str) -> Tuple[str, str]:
        """Generate URL for "Get Request State" API

        Args:
          id (str): ID of a request

        Returns:
          Tuple[str, str]: [URL, Method]
        """
        if not id:
            raise NoIdError("No ID provided for API")
        return self.api_string(f"request/{id}"), "GET"

    def api_restart(self, id: str) -> Tuple[str, str]:
        """Generate URL for "Restart Request" API

        Args:
          id (str): ID of a request

        Returns:
          Tuple[str, str]: [URL, Method]
        """
        if not id:
            raise NoIdError("No ID provided for API")
        return self.api_string(f"request/restart/{id}"), "POST"

    def api_cleanup(self, id: str) -> Tuple[str, str]:
        """Generate URL for "Cleanup Request" API

        Args:
          id (str): ID of a request

        Returns:
          Tuple[str, str]: [URL, Method]
        """
        if not id:
            raise NoIdError("No ID provided for API")
        return self.api_string(f"request/cleanup/{id}"), "POST"

    def api_extend(self, id: str, minutes: int) -> Tuple[str, str]:
        """Generate URL to extend request by `minutes` minutes

        Args:
            id (str): ID of a request
            minutes (int): amount of minutes

        Raises:
            NoIdError: Raised if no ID provided

        Returns:
            Tuple[str, str]: [URL, method]
        """
        if not id:
            raise NoIdError("No ID provided for API")
        return self.api_string(f"request/extend/{minutes}/{id}"), "POST"

    def api_force_cleanup(self, id: str) -> Tuple[str, str]:
        """Generate URL for "Force Cleanup Request" API

        Args:
          id (str): ID of a request

        Returns:
          Tuple[str, str]: [URL, Method]
        """
        if not id:
            raise NoIdError("No ID provided for API")
        return self.api_string(f"request/forcecleanup/{id}"), "POST"

    def drop_auth(self) -> None:
        """Resets Auth"""
        if self.auth and self.auth.params and "token" in self.auth.params:
            del self.auth.params["token"]

    def get_auth_token_pre(
            self,
            creds: str = "",
            datacenter: str = "") -> Union[str, Tuple[str, dict[str, str], dict[str, str]]]:
        if not self.auth:
            raise NoAuth("No Auth defined")

        logger.debug("Getting auth")
        if self.auth and self.auth.params and "token" in self.auth.params:
            return self.auth.params["token"]

        logger.debug("Requesting auth token")
        api = self.api_string("auth/token")
        headers = {"Content-Type": "application/json"}
        if self.auth.type == IDACAuthType.DCLOUD_SESSION.name:
            data = {"token": creds, "datacenter": datacenter, "type": "dcloud"}
        elif self.auth.type == IDACAuthType.WORKER.name:
            client_id, secret = itemgetter("client_id", "secret")(self.auth.params)
            data = {"token": get_worker_hash(client_id, secret), "type": "worker"}
        else:
            raise NoAuth("Auth type has no authentication request process defined")
        logger.debug(f"With data: {data}")

        return api, headers, data

    def verify_auth_response(self, response: httpx.Response, use_as_auth: bool) -> str:
        logger.debug(f"Got response: {response.status_code} {response.reason_phrase}")
        if response.status_code == 404:
            # old controller, no support for auth/token API, ignore :)
            logger.debug("Looks like controller doesn't support authentication. Ignoring")
            if use_as_auth:
                self.auth.params["token"] = "empty_token"
            return "empty_token"
        response.raise_for_status()
        json_body = json.loads(response.text)
        if isinstance(json_body, dict) and json_body.get("token", None):
            if use_as_auth:
                logger.debug("Saving token for future authentications")
                self.auth.params["token"] = json_body["token"]
            return json_body["token"]
        raise NoAuthTokenInResponse("Didn't receive auth token from controller")



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Extra


class NewRequest(BaseModel, extra=Extra.allow):
    demo: Optional[str] = None
    owner: Optional[str] = None
    id: Optional[str] = None
    cloud: Optional[str] = None
    recipePath: Optional[str] = None
    recipeName: Optional[str] = None
    location: Optional[str] = None
    recipeMode: Optional[str] = None
    datacenter: Optional[str] = None



import platform
import os.path
from warnings import warn
import xmltodict
import json
from typing import Any, Dict, Literal, Optional, Union
from deepmerge import always_merger

from idac_sdk.errors import (
    BadXmlFileError,
    IncorrectSessionData,
    SessionDataFileNotFoundError,
    SessionDataFileNotFoundWarning,
)
from idac_sdk.models.new_request import NewRequest

DEFAULT_SESSION_XML_UNIX = "/dcloud/session.xml"
DEFAULT_SESSION_XML_WIN = "c:\\dcloud\\session.xml"


class SessionData:
    data: NewRequest

    def __init__(
        self,
        session_xml_path: Optional[Union[str, Literal[False]]] = None,
        initial_data: Union[NewRequest, Dict[str, Any], None] = None,
        **kwargs,
    ) -> None:
        """SessionData Object

        Handles data of a session.

        Args:
            session_xml_path (Union[str, bool], optional): Path to session.xml file. If set to None
                will try to load either `c:\\dcloud\\session.xml` if on Windows or
                `/dcloud/session.xml` otherwise. If set to False will not load anything.
                Defaults to None.
            initial_data (Union[NewRequest, Dict[str, Any]], optional): Initial session data.
                If both session_xml_path and initial_data provided, data will be merged and values
                from initial_data will overwrite values from session.xml. Defaults to None.
            **kwargs: Any other named parameter will be considered as a parameter which should be
                added to data

        Raises:
            IncorrectSessionData: if incorrect `initial_data` provided

        Returns:
            None
        """
        if session_xml_path is not False:
            self.load_session_xml(file=session_xml_path)

        if initial_data or kwargs:
            if initial_data and (not isinstance(initial_data, NewRequest) and
                                 not isinstance(initial_data, dict)):
                raise IncorrectSessionData("Incorrect initial data provided")

            # ensure that d is dictionary
            if initial_data:
                d = (initial_data.dict(exclude_unset=True, exclude_none=True) if isinstance(
                    initial_data, NewRequest) else initial_data)
            else:
                d = dict()

            # all other params considered part of initial data as well, merge
            if kwargs:
                d = always_merger.merge(d, kwargs)

            if hasattr(self, "data"):
                # need to merge data
                self.data = NewRequest(**always_merger.merge(
                    self.data.dict(exclude_unset=True, exclude_none=True),
                    d,
                ))
            else:
                self.data = NewRequest(**d)
        else:
            if not hasattr(self, "data"):
                # create empty if not loaded from session.xml
                self.data = NewRequest()

    def load_session_xml(self,
                         file: Optional[Union[str, Literal[False]]] = None,
                         merge: bool = True) -> None:
        """Loads data from a file.
        If file is not specified loads from `c:\\dcloud\\session.xml` if on Windows or from
            `/dcloud/session.xml`

        Args:
            file (str, optional): File to load. Defaults to None
            merge (bool, optional): Tells if loaded data should be merged with existent or not.
                If False, existing data will be replace with loaded. Defaults to True

        Raises:
            SessionDataFileNotFoundError: If file doesn't exist

        Returns:
            None
        """
        xml_guessed = False
        if not file:
            xml_guessed = True
            if platform.system().lower() == "windows":
                file = DEFAULT_SESSION_XML_WIN
            else:
                file = DEFAULT_SESSION_XML_UNIX

        if not os.path.exists(file):
            if not xml_guessed:
                raise SessionDataFileNotFoundError(f"File {file} not found.")
            else:
                warn(f"session.xml not found by path '{file}'.", SessionDataFileNotFoundWarning)
                return

        d = None
        with open(file, "r") as file_handle:
            d = xmltodict.parse(file_handle.read())
            d = json.loads(json.dumps(d))

        if isinstance(d, dict):
            if "session" in d:
                new_data = NewRequest(**d["session"])
            else:
                new_data = NewRequest(**d)

            if hasattr(self, "data"):
                self.data = (NewRequest(**always_merger.merge(
                    self.data.dict(exclude_unset=True, exclude_none=True),
                    new_data.dict(exclude_unset=True, exclude_none=True),
                )) if merge else new_data)
            else:
                self.data = new_data
        else:
            raise BadXmlFileError(f"File {file} is not a useful XML file")

    def set(self, key: str, value: Any) -> None:
        """Adds or updates key/value to session data

        Args:
            key (str): Key to add to the session data
            value (Any): Value of the key

        Returns:
            None
        """
        setattr(self.data, key, value)

    def has(self, key: str) -> bool:
        """Check if `key` exists in session data

        Args:
            key (str): Key name to check

        Returns:
            bool: True if key exists, False otherwise
        """
        return key in self.data.__fields_set__

    def get(self, key: str) -> Any:
        """Returns value of a key. Throws if key doesn't exist

        Args:
            key (str): Key name to get

        Returns:
            Any: Value of the key
        """
        return getattr(self.data, key)

    def delete(self, key: str) -> None:
        """Delete a key from session data

        Args:
            key (str): Key name to delete

        Returns:
            None
        """
        delattr(self.data, key)

    def dict(self, *args, **kwargs) -> dict:
        """Generates dictionary from data

        Returns:
            dict: Data as a dictionary
        """
        return self.data.dict(*args, **kwargs)



import json
import platform
from typing import List
import asyncclick as click
import logging
from httpx import URL

from idac_sdk.log import logger
from idac_sdk import SessionData
from idac_sdk.models.request_state import RequestState

_debug_options = [
    click.option("--debug", default=False, is_flag=True, help="Enable debug", hidden=True)
]


def set_logger_level(debug: bool) -> None:
    if debug:
        logger.setLevel(logging.DEBUG)
        logger.debug("Debug level enabled")
    else:
        logger.setLevel(logging.WARN)


def add_options(options):

    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def filter_dict(dictObj, callback):
    newDict = dict()
    # Iterate over all the items in dictionary
    for (key, value) in dictObj.items():
        # Check if item satisfies the given condition then add to new dict
        if callback((key, value)):
            newDict[key] = value
    return newDict


def parse_data(ctx, param, raw: List[str]):
    parsed = dict()
    if raw is not None and len(raw) > 0:
        for v in raw:
            if "=" not in v:
                raise click.BadParameter(message="Data should have format `KEY=VALUE`",
                                         ctx=ctx,
                                         param=param)
            key, value = v.split("=")
            parsed[key] = value if value else ""
        return parsed

    return None


def parse_json(ctx, param, raw: str):
    if raw is not None:
        return json.loads(raw)


DEFAULT_OUTPUT_UNIX = r"/dcloud/idac_sdk.out"
DEFAULT_OUTPUT_WIN = r"c:\dcloud\idac_sdk.out"


def default_output(ctx, param, raw: str):
    if raw:
        return raw

    if platform.system().lower() == "windows":
        return DEFAULT_OUTPUT_WIN
    else:
        return DEFAULT_OUTPUT_UNIX


def update_url_with_creds(url: str, data: SessionData) -> str:
    u = URL(url)
    qp = u.params
    qp = qp.set("dcloudCreds", data.get("creds"))
    if data.has("datacenter"):
        qp = qp.set("datacenter", data.get("datacenter"))
    u = u.copy_merge_params(qp)
    return f"{u}"


URL_ATTRS = ["loaderUrl", "outputUrl", "mLoaderUrl", "errorUrl"]


def append_creds(state: RequestState, data: SessionData) -> RequestState:
    d = state.dict(exclude_unset=True)
    for attr in URL_ATTRS:
        if attr in d and d[attr]:
            d[attr] = update_url_with_creds(d[attr], data)

    return RequestState(**d)


import asyncclick as click
from jsonpath_ng import parse

from idac_sdk import IDACRequestAsync
from idac_sdk.cli.lib._controller_options import _controller_options, with_controller
from idac_sdk.cli.lib._helpers import (
    add_options,
    _debug_options,
    set_logger_level,
)


@click.group()
def commands():
    pass


@commands.command(short_help="get state of a request")
@add_options(_controller_options)
@add_options(_debug_options)
@click.option(
    "--out-path",
    multiple=True,
    help="Extract value from response by JSON path, Multiple allowed",
    type=str,
    default=None,
)
@click.argument("request_id")
@with_controller
async def state(controller, request_id, out_path, **kwargs):
    """Get JSON state object of a request

    REQUEST_ID - id/uuid of the request"""
    set_logger_level(kwargs.get("debug", False))

    req = IDACRequestAsync(uuid=request_id, controller=controller)
    state = await req.get_state()
    if out_path:
        state_dict = state.dict()
        for o in out_path:
            jsonpath_expr = parse(o)
            for match in jsonpath_expr.find(state_dict):
                click.echo(match.value)
    else:
        click.echo(state.json())


@commands.command(short_help="get status of a request")
@add_options(_controller_options)
@add_options(_debug_options)
@click.argument("request_id")
@with_controller
async def status(controller, request_id, **kwargs):
    """Get current status of a request

    REQUEST_ID - id/uuid of the request"""
    set_logger_level(kwargs.get("debug", False))

    req = IDACRequestAsync(uuid=request_id, controller=controller)
    state = await req.get_state()
    click.echo(state.status)



import os
import re

from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def get_version():
    VERSIONFILE = "idac_sdk/_version.py"
    verstrline = open(VERSIONFILE, "rt").read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
        return verstr
    else:
        raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


setup(
    name="idac_sdk",
    version=get_version(),
    author="Cisco GDE CAT",
    author_email="cat-dev-support@cisco.com",
    description="Python 3 SDK for iDAC",
    license="MIT",
    keywords="idac sdk",
    url="https://www-github.cisco.com/GDE-Content-Engineering/idac-sdk-python",
    packages=find_packages(include=["idac_sdk", "idac_sdk.*", "tests"]),
    long_description=read("README.md"),
    install_requires=[
        "asyncclick>=8.0.3.2",
        "httpx>=0.22.0",
        "xmltodict>=0.12.0",
        "pydantic>=1.9.0,<2.0.0",
        "deepmerge>=1.0.1",
        "PyYAML>=5.4.1",
        "jsonpath-ng>=1.5.3",
    ],
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    entry_points={"console_scripts": ["idac = idac_sdk.cli.idac:main"]},
)



import os
from typing import Any
import warnings
import pydantic
import yaml
import idac_sdk
from idac_sdk.errors import OldConfigFoundException, OldConfigFoundWarning
from idac_sdk.models.config import iDACConfig


DIR_PATH = os.path.dirname(idac_sdk.__file__)
IDAC_CONFIG_FOLDER = os.path.join(DIR_PATH, ".idac")
IDAC_CONFIG_FILE = os.path.join(IDAC_CONFIG_FOLDER, "config")
IDAC_CONFIG_VERSION = "1.0"
EMPTY_CONFIG: Any = {
    "version": IDAC_CONFIG_VERSION,
    "defaults": {
        "idac_fqdn": "",
        "idac_proto": "https",
        "auth": {"type": idac_sdk.IDACAuthType.DCLOUD_SESSION.name, "params": {}},
        "api_version": "2.0",
    },
}


def have_config() -> bool:
    """Checks if config file exists

    Returns:
        bool: True if exists
    """
    return os.path.exists(IDAC_CONFIG_FILE)


def load_config(raise_error: bool = True) -> iDACConfig:
    """Loads either existing or empty iDAC config

    Returns:
        iDACConfig: loaded config
    """
    if not os.path.exists(IDAC_CONFIG_FOLDER):
        os.mkdir(IDAC_CONFIG_FOLDER)

    if os.path.exists(IDAC_CONFIG_FILE):
        with open(IDAC_CONFIG_FILE, "r") as file:
            current_config = yaml.safe_load(file)
            if not "version" in current_config or current_config["version"] != IDAC_CONFIG_VERSION:
                if raise_error:
                    raise OldConfigFoundException(
                        "Found an old version of the config file, please run 'idac config' again."
                    )
                warnings.warn("Old iDAC config found, ignoring current file", OldConfigFoundWarning)
                return iDACConfig(**EMPTY_CONFIG)
        try:
            result = iDACConfig(**current_config)
            return result
        except pydantic.error_wrappers.ValidationError:
            if raise_error:
                raise OldConfigFoundException(
                    "Found an invalid config file, please run 'idac config' again."
                )
            warnings.warn(
                "Found an invalid config file, ignoring current file", OldConfigFoundWarning
            )
            return iDACConfig(**EMPTY_CONFIG)
    else:
        return iDACConfig(**EMPTY_CONFIG)
